package dfki.protege.flora_tab;

import java.util.*;
import java.io.*;
import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.ui.*;


class ExportFlora {

  static String floraAxiomSlotPrefix = "fl_axiom";
  static String floraQuerySlot = "fl_query";
  static String floraResultVar = "Result";

  static void export(Project project, String fileName) {

    try {
    PrintWriter pw = new PrintWriter(new FileWriter(fileName));

    KnowledgeBase kb = project.getKnowledgeBase();
    String kbName = kb.getName();

    pw.println("% automatically generated from Protege KB '" + kbName + "'");
    pw.println();

    Vector unionClasses = new Vector(); // [ [c1,..,cn], .... ]

    Collection instances = kb.getInstances();
    exportFrames(pw, instances, unionClasses);

    exportUnionClasses(pw, unionClasses);

    exportAxioms(pw, instances);

    pw.flush();
    pw.close();

    } catch (Exception e) {
      e.printStackTrace();
    }

  }


  // export frames

  static void exportFrames(PrintWriter pw, Collection instances, 
                           Vector unionClasses) {
    pw.println("% frames\n");
    for (Iterator instanceIterator = instances.iterator(); 
         instanceIterator.hasNext();)
      exportFrame(pw, (Instance)instanceIterator.next(), unionClasses);
  }


  static void exportFrame(PrintWriter pw, Instance instance, 
                          Vector unionClasses) {
    if (!instance.isSystem()) {
      exportInstance(pw, instance);
      if (instance instanceof Cls)
	exportClass(pw, (Cls)instance, unionClasses);
    }
  }


  static void exportInstance(PrintWriter pw, Instance instance) {

    /*
    if (instance instanceof Slot) // not interesting ???
      return;
    */

    if (instance instanceof Facet) // not interesting ???
      return;

    /*
    if (instance instanceof Cls) {
      Cls type = instance.getDirectType();
      if (type != null && type.isDefaultClsMetaCls())
        // type is :STANDARD_CLASS
        // so there are no interesting own slots;
        // superclasses and template slots will be exported
        // with exportClass
        return;
    }
    */

    // f:t[own-slots].

    pw.print(symbol(instance.getName()));
    Cls type = instance.getDirectType();
    if (type != null)
      pw.print(":" + symbol(type.getName()));
    Collection ownSlots = instance.getOwnSlots();
    pw.print("[");
    boolean semicolonNeeded = false;
    for (Iterator ownSlotIterator = ownSlots.iterator();
      ownSlotIterator.hasNext();) {
      if (exportOwnSlot(pw, semicolonNeeded, 
                        instance, (Slot)ownSlotIterator.next()))
        semicolonNeeded = true;
    }
    pw.println("].");

  }


  static boolean exportOwnSlot(PrintWriter pw, boolean semicolonNeeded,
      Instance instance, Slot ownSlot) {

    if (ownSlot.isSystem())
      // we are not interested in :NAME, :DIRECT-TYPE etc.
      return false;

    if (instance.getOwnSlotValueCount(ownSlot) > 0) {

      if (semicolonNeeded)
        pw.print(";\n    "); // COMMA
      ValueType valueType = instance.getOwnSlotValueType(ownSlot);
      boolean multiple = instance.getOwnSlotAllowsMultipleValues(ownSlot);
      pw.print(symbol(ownSlot.getName()));
      pw.print("->");
      if (multiple) {
        pw.print(">");
        Collection values = instance.getOwnSlotValues(ownSlot);
        pw.print(values(valueType, values, true));
      } else {
        Object value = instance.getOwnSlotValue(ownSlot);
        pw.print(value(valueType, value));
      }

      return true;

    } else

      return false;

  }


  static void exportClass(PrintWriter pw, Cls cls, Vector unionClasses) {

    Collection superClasses = cls.getDirectSuperclasses();

    if (superClasses.isEmpty())
      return; // should never happen

    /* already checked
    if (cls.isSystem()) // we are not interested in system classes
      return;
    */

    for (Iterator superClassesIterator = superClasses.iterator();
         superClassesIterator.hasNext();) {
     
      Cls superClass = (Cls)superClassesIterator.next();
      pw.print(symbol(cls.getName()) + "::" + symbol(superClass.getName()));
      if (!superClassesIterator.hasNext()) { 
        // show template slots as last molecule
        pw.print("[");
        Collection templateSlots = cls.getDirectTemplateSlots();
        for (Iterator templateSlotIterator = templateSlots.iterator();
             templateSlotIterator.hasNext();) {
          exportTemplateSlot(pw, cls, (Slot)templateSlotIterator.next(),
            unionClasses);
          if (templateSlotIterator.hasNext())
            pw.print(";\n    "); // COMMA
        }
        pw.print("]");
      }
      pw.println(".");

    }

  }


  static void exportTemplateSlot(PrintWriter pw, Cls cls, Slot slot, 
                                 Vector unionClasses) {
    pw.print(symbol(slot.getName()) + "=>");
    boolean multiple = cls.getTemplateSlotAllowsMultipleValues(slot);
    if (multiple)
      pw.print(">");
    pw.print(type(cls, slot, unionClasses));
  }


  // export union classes

  static void exportUnionClasses(PrintWriter pw, Collection unionClasses) {
    pw.println("\n\n% union classes\n");
    for (Iterator unionClassIterator = unionClasses.iterator();
         unionClassIterator.hasNext();) {
      Collection unionClass = (Collection)unionClassIterator.next();
      exportUnionClass(pw, unionClass);
    }
  }

  static void exportUnionClass(PrintWriter pw, Collection unionClass) {
    String unionClassName = unionClassName(unionClass);
    pw.println("% " + unionClassName);
    for (Iterator clsIterator = unionClass.iterator();
         clsIterator.hasNext();) {
      Cls cls = (Cls)clsIterator.next();
      pw.println(symbol(cls.getName()) + "::" + symbol(unionClassName) + ".");
    }
    Vector handledSuperClasses = new Vector();
    for (Iterator clsIterator = unionClass.iterator();
         clsIterator.hasNext();) {
      Cls cls = (Cls)clsIterator.next();
      Collection superClasses = cls.getDirectSuperclasses();
      for (Iterator superClsIterator = superClasses.iterator();
           superClsIterator.hasNext();) {
        Cls superCls = (Cls)superClsIterator.next();
        if (!handledSuperClasses.contains(superCls)) {
          pw.println(symbol(unionClassName) + "::" + 
                     symbol(superCls.getName()) + ".");
          handledSuperClasses.addElement(superCls);
        }
      }
    }
    pw.println();
  }


  // export axioms

  static void exportAxioms(PrintWriter pw, Collection instances) {
    pw.println("\n\n% axioms\n");
    for (Iterator instanceIterator = instances.iterator(); 
         instanceIterator.hasNext();) {
      Instance instance = (Instance)instanceIterator.next();
      Collection ownSlots = instance.getOwnSlots();
      for (Iterator ownSlotIterator = ownSlots.iterator();
           ownSlotIterator.hasNext();) {
        Slot ownSlot = (Slot)ownSlotIterator.next();
        String ownSlotName = ownSlot.getName();
        if (ownSlotName.startsWith(floraAxiomSlotPrefix)) {
          ValueType valueType = instance.getOwnSlotValueType(ownSlot);
          boolean multiple = instance.getOwnSlotAllowsMultipleValues(ownSlot);
          if (!multiple && valueType.equals(ValueType.STRING)) {
            Object value = instance.getOwnSlotValue(ownSlot);
            if (value != null) {
              String axiom = value.toString().trim();
              if (!axiom.equals("")) {
                pw.println("% " + instance.getName() + "." + ownSlotName);
                pw.println(axiom);
                pw.println();
              }
            }
          } else { // should never happen
            System.out.println("axiom slot should be of type single STRING: " 
                               + instance);
          }
        }
      }
    }
  }


  // utils

  static String symbol(String string) {
    if (string == null || string.equals(""))
      return "''"; // should never happen
    else {
      char first = string.charAt(0);
      if (Character.isLowerCase(first)) {
        boolean onlyGoodChars = true;
        int len = string.length();
        for (int i = 0; i < len; i++) {
          if (!goodChar(string.charAt(i)))
            onlyGoodChars = false;
        }
        if (onlyGoodChars)
          return string;
        else
          return "'" + normalize(string) + "'";
      } else // :THING etc. -> ???
        return "'" + normalize(string) + "'";
    }
  }

  static boolean goodChar(char c) {
    return Character.isLetterOrDigit(c) || c == '_';
  }

  static String normalize(String string) {
    StringBuffer buffer = new StringBuffer();
    int len = string.length();
    for (int i = 0; i < len; i++) {
      char c = string.charAt(i);
      if (c == '\'')
        buffer.append("''");
      else if (c == '\\')
        buffer.append("\\\\");
      else if (c == '\n')
        buffer.append("\\n");
      // else ... \t etc.
      else 
        buffer.append(c);
      /* don't activate this -- normalize is also used for normalizing
         Flora file names !!!
      if (i > 100) { // :-( !!!
        buffer.append(" ...");
        break;
      }
      */
    }
    return buffer.toString();
  }


  static String values(ValueType valueType, Collection values, 
                       boolean withBraces) {
    StringBuffer buffer = new StringBuffer();
    if (withBraces)
      buffer.append("{");
    for (Iterator valueIterator = values.iterator();
         valueIterator.hasNext();) {
      buffer.append(value(valueType, valueIterator.next()));
      if (valueIterator.hasNext())
        buffer.append(",");
    }
    if (withBraces)
      buffer.append("}");
    return buffer.toString();
  }

  static String value(ValueType valueType, Object value) {
    // this version ignores valueType!
    if (value == null)
      return null; // ???
    else if (value instanceof Frame) // CLS, INSTANCE
      return symbol(((Frame)value).getName());
    else if (value instanceof Number) // FLOAT, INTEGER
      return value.toString();
    else // BOOLEAN, STRING, SYMBOL
      return symbol(value.toString());
  }


  /* old, incomplete version
  static String value(ValueType valueType, Object value) {
    if (valueType == ValueType.ANY) {
      return symbol(value.toString()); // incomplete !
    } else if (valueType == ValueType.BOOLEAN) {
      return symbol(value.toString());
    } else if (valueType == ValueType.CLS) {
      return symbol(value.toString());
    } else if (valueType == ValueType.FLOAT) {
      return value.toString();
    } else if (valueType == ValueType.INSTANCE) {
      return symbol(((Instance)value).getName());
    } else if (valueType == ValueType.INTEGER) {
      return value.toString();
    } else if (valueType == ValueType.STRING) {
      return symbol(value.toString());
    } else if (valueType == ValueType.SYMBOL) {
      return symbol(value.toString());
    } else {
      System.out.println("Warning: unknown value type " + valueType);
      return symbol(value.toString());
    }
  }
  */


  static String type(Cls cls, Slot slot, Vector unionClasses) {
    ValueType valueType = cls.getTemplateSlotValueType(slot);
    if (valueType == ValueType.ANY) {
      return "object"; // ???
    } else if (valueType == ValueType.BOOLEAN) {
      return "boolean"; // ???
    } else if (valueType == ValueType.CLS) {
      return "class"; // ???
    } else if (valueType == ValueType.FLOAT) {
      return "float";
    } else if (valueType == ValueType.INSTANCE) {
      Collection allowedClasses = cls.getTemplateSlotAllowedClses(slot);
      if (allowedClasses.isEmpty())
        return "bottom"; // F-Logic has no type checkin, so what ...
      else
        if (allowedClasses.size() == 1)
          return symbol(((Cls)allowedClasses.iterator().next()).getName());
        else
          return symbol(unionClass(allowedClasses, unionClasses));
    } else if (valueType == ValueType.INTEGER) {
      return "integer";
    } else if (valueType == ValueType.STRING) {
      return "string";
    } else if (valueType == ValueType.SYMBOL) {
      return "symbol";
    } else {
      System.out.println("Warning: unknown value type " + valueType);
      return symbol(slot.getName().toString());
    }
  }

  static String unionClass(Collection allowedClasses, Vector unionClasses) {
    // sort
    Vector sortetAllowedClasses = new Vector(allowedClasses);
    Collections.sort(sortetAllowedClasses, new FrameComparator());
    // add to unionClasses if new
    if (!unionClasses.contains(sortetAllowedClasses))
      unionClasses.addElement(sortetAllowedClasses);
    // create name
    return unionClassName(sortetAllowedClasses);
  }

  static String unionClassName(Collection unionClass) {
    StringBuffer unionClassName = new StringBuffer();
    for (Iterator clsIterator = unionClass.iterator();
         clsIterator.hasNext();) {
      Cls cls = (Cls)clsIterator.next();
      unionClassName.append(cls.getName());
      if (clsIterator.hasNext())
        unionClassName.append("_or_");
    }
    return unionClassName.toString();
  }


  static Collection getResultValues(String result) {
    Vector values = new Vector();
    StringTokenizer lines = new StringTokenizer(result, "\n");
    String resultStart = floraResultVar + " = ";
    int len = resultStart.length();
    while (lines.hasMoreTokens()) {
      String line = lines.nextToken();
      if (line.startsWith(resultStart)) {
        String value = line.substring(len).trim();
        values.addElement(value);
      }
    }
    return values;
  }


}


